// --- [ USI_Slave.c ] --------------------------------------------------------
//
//       tab = 3 - 24.02.2012

// This file has been prepared for Doxygen automatic documentation generation.
/*! \file ********************************************************************
*
* Atmel Corporation ### GENDERTE VERSION ###
*
* File              : USI_Slave.c
* Compiler          : IAR EWAAVR 4.11A
* Revision          : $Revision: 1.14 $
* Date              : $Date: Friday, December 09, 2005 17:25:38 UTC $
* Updated by        : $Author: jtyssoe $
*
* Supported devices : All device with USI module can be used.
*
* AppNote           : AVR312 - Using the USI module as a I2C slave
*
* Description       : Functions for USI_USI_receiver and USI_USI_transmitter.
*
* NDERUNG:  Getrennte Buffer fr Senden und Empfangen
*            zustzliches USI_state (0) eingefgt
*            zustzliche Abfrage, ob Daten empfangen wurden
*            nicht bentigte Funktionen entfernt,
*				 bei einem TWI-Start (+read)  wird USI_TxHead zurckgesetzt,
*            bei einem TWI-Start (+write) wird USI_RxHead zurckgesetzt
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "USI_Slave.h"  // genderte Include-Datei


uint8_t USI_RxBuf[USI_RX_BUFFER_SIZE];
uint8_t USI_TxBuf[USI_TX_BUFFER_SIZE];
volatile uint8_t USI_State;
volatile uint8_t USI_RxHead;
volatile uint8_t USI_TxHead;

// ---------------------------------------------------------------------------
void USI_SLA_Init( void )
{

USI_State = USI_SLAVE_WAITING;

PORT_USI |=  (1<<PORT_USI_SCL);                   // Set SCL high
PORT_USI |=  (1<<PORT_USI_SDA);                   // Set SDA high
DDR_USI  |=  (1<<PORT_USI_SCL);                   // Set SCL as output
DDR_USI  &= ~(1<<PORT_USI_SDA);                   // Set SDA as input
USICR     =  (1<<USISIE)|(0<<USIOIE)|             // Enable Start Condition Interrupt. Disable Overflow Interrupt.
             (1<<USIWM1)|(0<<USIWM0)|             // Set USI in Two-wire mode. No USI Counter overflow prior
                                                  // to first Start Condition (potentail failure)
             (1<<USICS1)|(0<<USICS0)|(0<<USICLK)| // Shift Register Clock Source = External, positive edge
             (0<<USITC);
USISR    = 0xF0;                                  // Clear all flags and reset overflow counter
}

// ---------------------------------------------------------------------------
// solange USI_State > 0 ist der Slave beschftigt - sofern die bertragung
// vom Master mit NACK und anschlieendem STOP beendet wird.
// sendet er aber kein NACK, sondern direkt ein STOP, dann hat der Slave keine Mglichkeit,
// dies mittels Interrupt zu erkennen (STOP setzt nur ein Flag, lst aber keinen Interrupt aus).
// daher muss das Flag USIPF geprft werden.
/*
uint8_t USI_SLA_Busy( void )
{
	return (( USI_State) && (!(USISR & (1<<USIPF))));
}
*/

uint8_t USI_SLA_Rx_Cnt( void )
{
// Wenn eine bertragung im Gange ist, dann wartet die Routine hier bis zum
// Abschluss.
// Alternativ: FALSE zurckgeben, whrend eine bertragung im Gange ist.
// if (( USI_State) && (!(USISR & (1<<USIPF)))) return(0);
//
// Aber das bringt zustzliche Probleme bei einem Sleep().
// Also nehmen wir die (kurze Wartezeit) in Kauf.
	while ( USI_State && (!(USISR & (1<<USIPF))));

	return (USI_RxHead);
}

// ---------------------------------------------------------------------------
// Holt die empfangenen Daten ab,
// Die Anzahl steht in USI_RxHead
// vorher muss mit USI_USI_Rx_Cnt() festgestellt worden sein, dass Daten vorhanden sind!

void USI_SLA_Get_Data(uint8_t * data)
{
uint8_t i = 0;
do {
	data[i] = USI_RxBuf[i];
	}
	while (++i < USI_RxHead);

USI_RxHead = 0;
}

// ---------------------------------------------------------------------------
// Kopiert Daten in den lokalen Sendepuffer zur Abholung durch den Master
void USI_SLA_Put_Data(uint8_t * data, uint8_t cnt)
{
uint8_t i;
if (cnt > USI_TX_BUFFER_SIZE) cnt = USI_TX_BUFFER_SIZE;
for (i = 0; i < cnt; i++) USI_TxBuf[i] = data[i];
}

// ---------------------------------------------------------------------------
// wird am Beginn einer USI-TWI-Transaktion aufgerufen
ISR(USI_START_VECTOR)
{
   unsigned char tmpUSISR;                            // Temporary variable to store volatile
   tmpUSISR = USISR;                                  // Not necessary, but prevents warnings
																		// Set default starting conditions for new TWI package
   USI_State = USI_SLAVE_CHECK_ADDRESS;
   DDR_USI  &= ~(1<<PORT_USI_SDA);                    // Set SDA as input
   while ( (PIN_USI & (1<<PORT_USI_SCL)) & !(tmpUSISR & (1<<USIPF)) );
																		// Wait for SCL to go low to ensure the "Start Condition" has completed.
                                                      // If a Stop condition arises then leave the interrupt to prevent waiting forever.
   USICR  = (1<<USISIE)|(1<<USIOIE)|                  // Enable Overflow and Start Condition Interrupt. (Keep StartCondInt to detect RESTART)
            (1<<USIWM1)|(1<<USIWM0)|                  // Set USI in Two-wire mode.
            (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|      // Shift Register Clock Source = External, positive edge
            (0<<USITC);
   USISR  = (1<<USI_START_COND_INT)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      // Clear flags
            (0x0<<USICNT0);                           // Set USI to sample 8 bits i.e. count 16 external pin toggles.
}

// ---------------------------------------------------------------------------
// wird whrend der USI-TWI-Transaktion aufgerufen, wenn 8Bit bertragen sind
ISR(USI_OVERFLOW_VECTOR)
{
  switch (USI_State)
  {
   // ---------- Address mode -------------------------------------------------

   // Check address and send ACK (and next USI_SLAVE_SEND_DATA) if OK, else reset USI.
    case USI_SLAVE_CHECK_ADDRESS:

      if ((USIDR & 0xFE) == USI_SLA_ADR)
			{
			if ( USIDR & 0x01 )		// wenn Bit.0 gesetzt ist, dann liest der Master vom Slave
				{
				USI_State = USI_SLAVE_SEND_DATA;
				USI_TxHead = 0;
				}

			else                    // der Master schreibt Daten auf den Slave
				{
				USI_State = USI_SLAVE_REQUEST_DATA;
				USI_RxBuf[0] = USIDR;
				USI_RxHead = 1;
				}

			SET_USI_TO_SEND_ACK();
			}											// beim Lesen des TX_Buffer

      else											// falsche Adresse - Beenden
			{
			USI_State = USI_SLAVE_WAITING; // Ergnzung
			SET_USI_TO_USI_START_CONDITION_MODE();
			}

		break;

   // ----- Slave Send Data Mode ---------------------------------------------

   // Check reply and goto USI_SLAVE_SEND_DATA if OK, else reset USI.
   case USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA:
      if ( USIDR ) // If NACK, the master does not want more data.
			{
			USI_State = USI_SLAVE_WAITING; // Ergnzung
			SET_USI_TO_USI_START_CONDITION_MODE();
			break;
			}

	// From here we just drop straight into USI_SLAVE_SEND_DATA if the master sent an ACK

   // Copy data from buffer to USIDR and set USI to shift byte. Next USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA
   case USI_SLAVE_SEND_DATA:
		USIDR = USI_TxBuf[USI_TxHead];
      USI_TxHead = ((USI_TxHead + 1) & USI_TX_BUFFER_MASK);
      USI_State = USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA;
      SET_USI_TO_SEND_DATA();
      break;

   // Set USI to sample reply from master. Next USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA
   case USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA:
      USI_State = USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA;
      SET_USI_TO_READ_ACK();
      break;

   // ----- Slave Read Data Mode ---------------------------------------------

   // Set USI to sample data from master. Next USI_SLAVE_GET_DATA_AND_SEND_ACK.
   case USI_SLAVE_REQUEST_DATA:
      USI_State = USI_SLAVE_GET_DATA_AND_SEND_ACK;
      SET_USI_TO_READ_DATA();
      break;

   // Copy data from USIDR and send ACK. Next USI_SLAVE_REQUEST_DATA
   case USI_SLAVE_GET_DATA_AND_SEND_ACK:
      USI_RxBuf[USI_RxHead]  = USIDR;
      USI_RxHead = ( USI_RxHead + 1 ) & USI_RX_BUFFER_MASK;
      USI_State = USI_SLAVE_REQUEST_DATA;
      SET_USI_TO_SEND_ACK();
      break;
	}
}

// --- [ eof ] ----------------------------------------------------------------
